/*
 *************************************************************************************************************
 * PickSortElement.h       
 *
 * Die Struktur, die das Bild und die Bildinformationen fuer die Anzeige kapselt.
 * Idee ist es, eine ausreichend flexible, doppelt verkette Liste parallel zu den im Grid gespeicherten
 * Grafikelementen zu fuehren. Mit den Klassenlementen, die die doppelt verkettete Liste erzeugen, wird
 * also sortiert, editiert, usf., mit den Grafikelementen kann man Drag-n-Drop-Operationen ausfuehren.
 *
 * Copyright (C) Andreas Eichendorf 2014 <PickSort@Andreas-Petra.de>            
 *************************************************************************************************************
 */

#ifndef PICKSORTELEMENT_H
#define PICKSORTELEMENT_H

#include <iostream>
#include <string>
#include <time.h>
#include <gtkmm.h>
#include <exiv2/image.hpp>
#include <exiv2/exif.hpp>

#include "PickSortGenerelles.h"


class ElementEventBoxClass : public Gtk::EventBox, public GenerellesClass
{
 public:
  // Konstruktor fuer die jeweils ersten und letzten, nur symbolischen Elemente in den Editierspalten
  // Wegen der Sortierung: ASCII-Anfang = " " (Blank); ASCII-Ende   = "~" (Tilde):
  ElementEventBoxClass(const std::string& Filename,
                       const std::string& Fotografierdatum,
                       bool Sichtbarkeit,
                       unsigned short SpaltenNrHier);
  // Konstruktor fuer die normalen Elemente:
  ElementEventBoxClass(const std::string& Verzeichnisname,
                       const std::string& Filename, 
                       const std::string& Fotografierdatum,
                       short Bewertung,
                       bool Sichtbarkeit,
                       Glib::RefPtr<Gdk::Pixbuf> Pixelbuffer,
                       unsigned short SpaltenNrHier,
                       int BildDrehung);
  virtual ~ElementEventBoxClass();  
  // Setter:
  void SetElementVerzeichnis(const std::string& VerzeichnisName);
  void SetElementFilename(const std::string& FileName);
  void ElementUeberschreibeVerzeichnisNeu(const std::string& VerzeichnisName);
  void ElementUeberschreibeFilenameNeu(const std::string& FileName);
  void SetElementBewertung(short Bewertung);
  void SetElementSichtbar(bool Sichtbarkeit);  
  void SetElementFileGespeichert();
  void SetElementBewertungGespeichert();
  void SetElementDrehungGespeichert();
  void SetElementUmbenannt(bool Umbenannt = true);
  void SetElementUmbewertet(bool Umbewertet = true);  
  // Getter:
  const std::string& GetElementVerzeichnisNeu();
  const std::string& GetElementVerzeichnisAlt();
  const std::string& GetElementFilenameNeu();
  const std::string& GetElementFilenameAlt();  
  const std::string& GetElementDatum();
  short GetElementBewertungNeu();
  short GetElementBewertungAlt();
  bool GetElementSichtbar();
  bool GetElementUmbenannt();
  bool GetElementUmbewertet(); 
  int GetElementDrehung();
  bool GetIstStartEndeElement();
  ElementEventBoxClass* GetElement1Gewaehlt();
  ElementEventBoxClass* GetElement2Gewaehlt();
  // Tests:
  bool ElementVerzeichnisGeaendert();
  bool ElementFilenameGeaendert();
  bool ElementBewertungGeaendert();
  bool ElementDrehungGeaendert();
  // Benutzerinteraktionen:
  void ElementRechtsDrehen();
  void ElementLinksDrehen();
  void ElementBetrachten();
  void ElementEditieren();
  // Zeiger fuer doppelt verkettete Liste. Diese wird hauptsaechlich von den Spalten verwendet.
  // Die VerkettungsZeiger sind public, so dass man schnelle Routinen damit schreiben kann:
  ElementEventBoxClass* ElementDavor;
  ElementEventBoxClass* ElementDanach; 
 protected:
  // Kinderwidgets:     
  Gtk::Image ElementImage;
  Gtk::Label ElementLabelFilename;
  Gtk::Label ElementLabelAufnahmedatum;
  Gtk::Grid ElementLabelGrid;
  Gtk::RadioButton ElementBewertungButton[6]; 
  Gtk::Grid ElementBewertungGrid;
  Gtk::Alignment ElementLabelAlignment;
  Gtk::Grid ElementGrid;
  // Signalhandler:
  // ... das ganze Element betreffend:
  bool ElementButtonPressHandler(GdkEventButton* ElementButtonPressEvent);
  bool ElementEnterNotifyHandler(GdkEventCrossing* ElementEnterNotifyEvent);
  bool ElementLeaveNotifyHandler(GdkEventCrossing* ElementLeaveNotifyEvent);
  // ... die Bewertungsbuttons betreffend:
  void ElementBewertungsButtonHandler(int ButtonNummer);
  // ...Drag und Drop betreffend:
  bool ElementDragDropHandler(const Glib::RefPtr<Gdk::DragContext>& Kontext, int X, int Y, guint Zeit);
 private:
  // Kontrollvariablen fuer Drag und Drop, Projektion, etc.:
  static ElementEventBoxClass* Element1Gewaehlt;
  static ElementEventBoxClass* Element2Gewaehlt;
  static unsigned short NrSpalteAktuellGewaehlt;
  unsigned short NrSpalteHier;
  static std::vector<Gtk::TargetEntry> TargetVektor;  
  static bool TargetVektorInitialisiert;
  // Zu speichernde Informationselemente:
  std::string VerzeichnisAlt;
  std::string VerzeichnisNeu;
  std::string FilenameAlt;
  std::string FilenameNeu;
  std::string Datum;
  short BewertungAlt;
  short BewertungNeu;
  int Drehung;
  bool Sichtbar;
  bool  StartEndeElement;
  // Kontrollvariablen fuer Lade- und Speicheroperationen in den Spalten:
  bool ElementUmbenannt;  
  bool ElementUmbewertet;
  // Kontrollvariable fuer Drag und Drop:
  bool ElementGewaehlt;
};  // class ElementEventBoxCLass


//************************************************************************************************************


// Managerklasse fuer die eigentliche - s.u. - Betrachter-Klasse:
class ElementBetrachtenWindowClassManager
{
 public:
  ElementBetrachtenWindowClassManager();
  virtual ~ElementBetrachtenWindowClassManager();
  void WindowOeffnen(ElementEventBoxClass& ZuZeigendesBild); 
};  // class ElementBetrachtenWindowClassManager


// Eine Klasse zum Betrachten einzelner Fotos, wird von den jeweiligen Bildelementen benutzt:
class ElementBetrachtenWindowClass : public Gtk::Window, public GenerellesClass
{
 public:
  // Allokation des Fensters geschieht automatisch ueber die friend-Klasse:
  friend class ElementBetrachtenWindowClassManager;
  virtual ~ElementBetrachtenWindowClass();    
 protected:
  // Kinderwidgets:
  Glib::RefPtr<Gdk::Pixbuf> ElementBetrachtenPixbufQuelle;
  Glib::RefPtr<Gdk::Pixbuf> ElementBetrachtenPixbufZwischen;  
  Glib::RefPtr<Gdk::Pixbuf> ElementBetrachtenPixbufZiel;
  // File und Verzeichnis im Window-Kopf geht schlecht, da der Fenstermanager seine Hand auf dem Aussehen
  // hat. Ich fuehre also noch eine Label-Zeile mit Verzeichnis und Filenamen mithilfe eines Grids ein:
  Gtk::Grid ElementBetrachtenGrid;
  Gtk::Label ElementBetrachtenLabel;
  Gtk::Image ElementBetrachtenImage;
  Gtk::Button ElementBetrachtenButtonVor;
  Gtk::Button ElementBetrachtenButtonZurueck;
  // Signalhandler zum Schliessen des Elementes:
  bool ElementBetrachtenSchliessenEvent(GdkEventAny* Event);
  // Callback-Funktion beim Aufziehen des Fensters:
  bool ElementBetrachtenKonfigureEvent(GdkEvent* Event);
  // Callback-Funktionen fuer die Vor- und Zurueck-Buttons:
  void ElementBetrachtenButtonVorClick();
  void ElementBetrachtenButtonZurueckClick();
 private:
  // Konstruktion nur ueber den friend-Klassenmanager. So hat der aufrufende Code nichts mit der
  // Allokation und Deallokation der Fensterobjekte zu schaffen: 
  ElementBetrachtenWindowClass(ElementEventBoxClass& Bild);
  // Blaettern durch eine Spalte von Bildern, wird von den Tastencallbacks aufgerufen:
  void ElementBetrachtenVorZurueckBlaettern();
  // Speicher fuer die Bildadresse, notwendig fuer das Weiterblaettern:
  ElementEventBoxClass* ZuZeigendesBild;
  // Breite und Hoehe des Rahmens um das Bild. Das ist der freie Einstellhub beim Verkleinern des Bildes,
  // d.h. je breiter, desto schneller laeuft die Verkleinerung:
  const int RahmenstaerkeHorizontal;
  const int RahmenstaerkeVertikal;
  // Seitenverhaeltnisse des Bildes und Laenge und Breite des Windows:
  float AspektQuelle;
  int BreiteWindowIst;
  int HoeheWindowIst;
  // Hoehe des Labels ueber dem eigentlichen Foto:
  int HoeheLabel;
};  // class ElementBetrachtenWindowClass


#endif  // PICKSORTELEMENT_H

/*
 * PickSortElement.h *****************************************************************************************
 */
