/*
 *************************************************************************************************************
 * PickSortProXOpenGL.cpp       
 *
 * OpenGL- und X11-Funktionen fuer das Projektorprogramm.
 *
 * Copyright (C) Andreas Eichendorf 2014 <PickSort@Andreas-Petra.de>            
 *************************************************************************************************************
 */


#include "PickSortProXOpenGL.h"


bool PickSortProXOpenClass::Initialisiert = false;


PickSortProXOpenClass::PickSortProXOpenClass()
{
 // Es gibt den Bildschirm nur ein einziges Mal in diesem Programm:
 if (Initialisiert)
  {
   return;
  }
 // Erzeuge die Verbindung zum X-Server mit der Bildschirmausgabe auf dem lokalen Rechner => (NULL):
 XSchnittstelle = XOpenDisplay(NULL); 
 if(XSchnittstelle == NULL)
  {
   std::cerr << "XOpen::XOpen: Unable to establish connection to X-server. => exit(1)" << std::endl;
   exit(1); 
  }
 // Einen Handler des Hauptfensters generieren (DefaultRootWindow ist ein Macro in der Xlib).
 // Der X-Server denkt in der Hierarchie von Fenstern, daher gibt es immer ein Root-Window:
 WurzelFenster = DefaultRootWindow(XSchnittstelle);
 // Eine Sammlung an moeglichen Attributlisten anlegen,
 // die unterschiedliche Anforderungen an die Grafikkarte stellen:
 // Es braucht das Programm keinen Tiefenpuffer. Ein Doppelpuffer muesste immer neu aufgebaut werden,
 // so dass man immer zwei Bilder uebereinander laden muesste, was extra Zeit kostet. Man kann ja auch mal
 // Grafikkarten haben, die nur 25 Bilder je Sekunde laden koennen. Daher ist es hier besser, mit nur einem
 // Puffer zu arbeiten, der nur immer neu ueberblendet wird:
 GLint AttributeZeiger1[] = {GLX_RGBA, GLX_DEPTH_SIZE,  0, None};
 // Generische Routine zum Ausprobieren verschiedener Attributkombinationen. Wird hier ja eigentlich nicht
 // gebraucht, weil die Anforderungen minimal sind:
 Attribute.push_back(AttributeZeiger1);   
 // Sichtfenstereinstellung, die am besten passt - es mussen mehrere getestet werden, um ggf. auch schwache
 // Grafikkarten bedienen zu koennen:
 for (std::vector<GLint*>::iterator AttributeIterator = Attribute.begin();
      AttributeIterator != Attribute.end();
      ++AttributeIterator)
  {
   // , 0, steht fuer die Bildschirmnummer 0:
   DarstellungsInfo = glXChooseVisual(XSchnittstelle, 0, *AttributeIterator);
   if(DarstellungsInfo != NULL)
    {
     break;
    }
  }
 if(DarstellungsInfo == NULL)
  {
   std::cerr << "XOpen::XOpen: Unable to find appropriate visual parameters. => exit(1)" << std::endl;
   exit(1);
  } 
 // Farbtafel der Anzeige erstellen:
 Farbtafel = XCreateColormap(XSchnittstelle, WurzelFenster, DarstellungsInfo->visual, AllocNone); 
 if (Farbtafel == 0)
  {
   std::cerr << "XOpen::XOpen: Unable to set up colormap. => exit(1)" << std::endl;
   exit(1);
  }
 // Initialiseren der XSetWindowAttributes-Struktur. Die Daten werden dann vom X-Server ausgewertet:
 XAttributeSet.colormap = Farbtafel; 
 // Auf diese Events wird das Fenster dann reagieren:
 XAttributeSet.event_mask = KeyPressMask | ExposureMask; 
 // Hier geht es nur um die maximale Laenge und Breite des Bildschirms mit der aktuellen Aufloesung,
 // die Routine holt die Daten des WurzelFensters:
 XGetWindowAttributes(XSchnittstelle, WurzelFenster, &WurzelfensterAttribute); 
 // Daten des Vollbildes ermitteln:
 BildschirmBreite = WurzelfensterAttribute.width;
 BildschirmHoehe = WurzelfensterAttribute.height;
 BildschirmSeitenverhaeltnis = static_cast<double>(BildschirmBreite) / static_cast<double>(BildschirmHoehe);
 // Jetzt kann das eigentliche Top-Level-Fenster angelegt werden. Es ist zuerst noch nicht sichtbar.
 // (Wenn man ein Fenster als WurzelFenster kreiert, dann ist das Fenster ein Top-Level-Fenster)
 // Bei Top-Level-Windows werden die Angaben x, y und Rahmenbreite oft ignoriert. Das Fenster wird
 // vom Fenstermanager dekoriert. Verodert findet man die Parameter, die veraendert wurden,
 // und so vom X-System neu bewertet werden muessen (erinnere: OpenGL hat eine state machine):
 Fenster = XCreateWindow(XSchnittstelle, WurzelFenster,
                         0, 0, BildschirmBreite, BildschirmHoehe, 0,
                         DarstellungsInfo->depth, InputOutput, DarstellungsInfo->visual,
                         CWColormap | CWEventMask, &XAttributeSet); 
 // Den Cursor verstecken: Es wird ein unsichtbarer Cursor erzeugt.
 // Der Destruktor schaltet dann beim Zerstoeren des Objektes wieder zurueck:
 char LeerePixelzeile[] = {0, 0, 0, 0};
 QuelleMaske = XCreateBitmapFromData(XSchnittstelle, Fenster, LeerePixelzeile, 4, 4);
 XColor Schwarz;
 Schwarz.red = Schwarz.green = Schwarz.blue = 0;
 Cursor VersteckterCursor =
  XCreatePixmapCursor(XSchnittstelle, QuelleMaske, QuelleMaske, &Schwarz, &Schwarz, 0, 0);
 XDefineCursor(XSchnittstelle, Fenster, VersteckterCursor);
 // Fenster sichtbar schalten, dadurch dass es auf den XSchnittstelle gemappt wird:
 XMapWindow(XSchnittstelle, Fenster);
 // Das Fenster nun in den Vollbildmodus umschalten. Hierfuer wird ein NetWM-Hint benutzt. Das scheint die
 // aktuelle, bei modernen Fenstermanagern immer funktionierende Methode zu sein. Allerdings muss der
 // Fenstermanager komischerweise darauf nicht zwangsweise reagieren - es ist ja nur ein 'Hint':
 XEvent XNetWMEvent;
 Atom WMStatus = XInternAtom(XSchnittstelle, "_NET_WM_STATE", false);
 Atom WMVollbild = XInternAtom(XSchnittstelle, "_NET_WM_STATE_FULLSCREEN", false);
 // Fuelle die Struktur XNetWMEvent zuerst mit Nullen aus:  
 memset(&XNetWMEvent, 0, sizeof(XNetWMEvent));
 // Belege dann nur die Felder fuer die Anforderung des Vollbildes:
 XNetWMEvent.type = ClientMessage;
 XNetWMEvent.xclient.window = Fenster;
 XNetWMEvent.xclient.message_type = WMStatus;
 XNetWMEvent.xclient.format = 32;
 // 1 schaltet ein, 0 schaltet wieder aus:
 XNetWMEvent.xclient.data.l[0] = 1;
 XNetWMEvent.xclient.data.l[1] = WMVollbild;
 XNetWMEvent.xclient.data.l[2] = 0;
 // Den Hint nun an die XSchnittstelle schicken:        
 XSendEvent(XSchnittstelle, WurzelFenster, false, SubstructureRedirectMask, &XNetWMEvent);
 // Zur Sicherheit, dass das X-System auch wirklich alle Befehle abgearbeitet hat:
 XSync(XSchnittstelle, false); 
 // Zeichenbereich 'GL-Umgebung' erstellen und mit X-Kanal verbinden. GL_TRUE: Direktes rendern auf lokaler
 // Grafikkarte anfordern; NULL: Mit DisplayListen koennen statische Grafikdaten in der Grafikkarte
 // gespeichert und dann sehr schnell ausgelesen werden. Bei NULL wird eine Displayliste, die in diesem
 // Kontext erzeugt wurde, nicht mit anderen Kontexten geteilt:
 GLKontext = glXCreateContext(XSchnittstelle, DarstellungsInfo, NULL, GL_TRUE);
 if (GLKontext == NULL)
  {
   std::cerr << "XOpen::XOpen: Unable to create GL-context. => exit(1)" << std::endl;
   exit(1);
  }
 // GL-Kontext mit dem Fenster verbinden. Aktuelles Fenster wird dabei zum Top-Level-Fenster: 
 glXMakeCurrent(XSchnittstelle, Fenster, GLKontext);
 // Aktuelles Betrachtungsfenster dimensionieren. Gibt die affine Transformation von X und Y
 // im Intervall [-1, 1], das sind die normalisierten Gerätekoordinaten, in Fensterkoordianten an
 // (Zustandsvariable). Da aber bereits beim Einrichten des Fensters die Maximalkoodinaten vergeben werden,
 // eruebrigt sich hier eigentlich die explizite Angabe:
 glViewport(0, 0, BildschirmBreite, BildschirmHoehe);
 // Projektionsmatrixstack anwaehlen (Zustandsvariable). Legt Art der Perspektive (Punkt- / Parallel-) fest:
 glMatrixMode(GL_PROJECTION);
 // Laden der Einheitsmatrix auf den Projektionsmatrixstack der Grafikkarte:
 glLoadIdentity();
 // Parallelprojektion, also keine Perspektive, 6 Clipping-Planes in die Projektionsmatrix multiplizieren.
 // Kamera-Position wird damit bei (0,0,0) angenommen:
 glOrtho(-1., 1., -1., 1., -1., 1.);
 // Zustandsvariable Sicht- und Ausrichtungsmatrix-Stack anwaehlen:
 glMatrixMode(GL_MODELVIEW);
 // Uberblendfunktionalitaet modal definieren: 
 glBlendFunc(GL_ONE_MINUS_CONSTANT_ALPHA, GL_CONSTANT_ALPHA); 
 // Zustandsvariable der Ueberblendung setzen:
 glEnable(GL_BLEND);
 // Raster-Ursprung modal setzen, dabei sind Sicht- und Ausrichtungsmatrix zu beachten:
 glRasterPos2f(-1., -1.);
 // Resetten
 glClearColor(0.1, 0.1, 0.1, 1.0);
 glClear(GL_COLOR_BUFFER_BIT);
 glFinish();
 // RAM-Bereich fuer den Bildschirmspeicher reservieren:
 PickSortProSmartRAMClass BildschirmKopie(BildschirmHoehe, BildschirmBreite, 4);
 // Bildschirmspeicher fuer den Geschwindigkeitstest in einen RAM-Bereich kopieren: 
 glReadPixels(0, 0, BildschirmBreite, BildschirmHoehe, GL_RGBA, GL_UNSIGNED_BYTE,
              static_cast<void *>(BildschirmKopie.GetAnfang())); 
 // Frames-Frequenz ermitteln:
 PickSortProTimerClass UeberblendeZeit;
 glFinish();
 UeberblendeZeit.ZeitnahmeStartKurz();
 // RAM-Bereich kurz = 20 Mal in den Bildschirmspeicher kopieren und anzeigen:
 for (register int Zaehler = 0; Zaehler < 20; ++Zaehler)
  {
   // Konstante Farbe wegen deren alpha-Wert definieren; Farbe ist der 18%-Graukarte nachvollzogen ;-):
   glBlendColor(0.5, 0.5, 0.5, 1.0);
   // Hintergrund mit gewuenschtem alpha-Wert versehen:
   glDrawPixels(BildschirmBreite, BildschirmHoehe, GL_RGBA, GL_UNSIGNED_BYTE,
                static_cast<void *>(BildschirmKopie.GetAnfang()));
   // Flushen, so dass das Bild in akzeptabler Verzoegerung auch sichtbar wird:
   glFinish();
  }
 UeberblendeZeit.ZeitnahmeEndeKurz(20.0);
 glClearColor(0.0, 0.0, 0.0, 1.0);
 glClear(GL_COLOR_BUFFER_BIT);
 glReadPixels(0, 0, BildschirmBreite, BildschirmHoehe, GL_RGBA, GL_UNSIGNED_BYTE,
              static_cast<void *>(BildschirmKopie.GetAnfang()));
 glFinish();
 UeberblendeZeit.ZeitnahmeStartLang();
 // Dann nochmals 80 Mal wiederholen, um konstante Zeitanteile tilgen zu koennen:
 for (register int Zaehler = 0; Zaehler < 80; ++Zaehler)
  {
   // Konstante Farbe wegen deren alpha-Wert definieren; Farbe ist der 18%-Graukarte nachvollzogen ;-):
   glBlendColor(0.5, 0.5, 0.5, 1.0);
   // Hintergrund mit gewuenschtem alpha-Wert versehen:
   glDrawPixels(BildschirmBreite, BildschirmHoehe, GL_RGBA, GL_UNSIGNED_BYTE,
                static_cast<void *>(BildschirmKopie.GetAnfang()));
   // Flushen, so dass das Bild in akzeptabler Verzoegerung auch sichtbar wird:
   glFinish();
  }
 UeberblendeZeit.ZeitnahmeEndeLang(80.0);
 Initialisiert = true;
}  // PickSortProXOpenClass


PickSortProXOpenClass::~PickSortProXOpenClass()
{
 // Vom unsichtbaren Cursor formal wieder auf den Standard-Cursor zurueckschalten:
 XUndefineCursor(XSchnittstelle, Fenster);
 // Pixmaps bleiben im X11-Speicherbereich stecken, auch nach Programmende, daher aufraeumen:
 XFreePixmap(XSchnittstelle, QuelleMaske);
 // Loese den GL-Grafikkontext vom X-Server:
 glXMakeCurrent(XSchnittstelle, None, NULL);
 // Loesche dann den GL-Grafikontext: 
 glXDestroyContext(XSchnittstelle, GLKontext);
 // Schliesse Fenster mit allen Subfenstern, loest nebenbei einen Destroy-Event aus:
 XDestroyWindow(XSchnittstelle, Fenster);
 // Schalte den X-Server ab:
 XCloseDisplay(XSchnittstelle);
 Initialisiert = false;
}  // ~PickSortProXOpenClass


bool PickSortProXOpenClass::XTastaturAbfrage(XBEFEHLE* BenutzerWunsch)
{
 XEvent Event;
 if (XCheckWindowEvent(XSchnittstelle, Fenster, KeyPressMask, &Event))
  {
   KeySym TastenSymbol = XkbKeycodeToKeysym(XSchnittstelle, Event.xkey.keycode, 0, 0);
   switch (TastenSymbol)
    {
     case XK_Page_Down:
      *BenutzerWunsch = NAECHSTESBILD;
      return true;
     case XK_Page_Up:
      *BenutzerWunsch = VORHERIGESBILD; 
      return true;           
     case XK_Up:
      *BenutzerWunsch = LAENGEREDAUER;
      return true;      
     case XK_Down:
      *BenutzerWunsch = KUERZEREDAUER;
      return true;   
     case XK_Home:
      *BenutzerWunsch = ERSTESBILD;
      return true;
     case XK_Left:
      *BenutzerWunsch = PAUSE;
      return true;      
     case XK_Right:
      *BenutzerWunsch = WEITER;
      return true;      
     case XK_Escape:
      *BenutzerWunsch = ENDE;
      return true;      
     default:
      *BenutzerWunsch = FALSCHETASTE;
      return false;
    }
  }
 else
  {
   *BenutzerWunsch = FALSCHETASTE;
   return false;
  }
}  // XTastaturAbfrage


int PickSortProXOpenClass::GetBildschirmBreite()
{
 return BildschirmBreite;
}  // GetBildschirmBreite


int PickSortProXOpenClass::GetBildschirmHoehe()
{
 return BildschirmHoehe;
}  // GetBildschirmHoehe


void PickSortProXOpenClass::XBildschirmUeberblenden(Magick::Blob NeuesBild)
{
 PickSortProTimerClass Timer;
 // + 1 wegen der einfacheren Schleifensteuerung:
 unsigned int GesamtZyklen = Timer.GetZyklenGesamtUeberblendung() + 1;
 register unsigned int Zaehler, ZaehlerLetzt;
 for (Zaehler = GesamtZyklen, ZaehlerLetzt = GesamtZyklen; Zaehler; --Zaehler)
  {
   // Konstante Farbe nur wegen deren alpha-Wert definieren; Farbe ist der 18%-Graukarte nachvollzogen ;-):
   glBlendColor(0.5, 0.5, 0.5, static_cast<float>(Zaehler - 1) / static_cast<float>(ZaehlerLetzt - 1));
   ZaehlerLetzt = Zaehler;
   // Hintergrund mit gewuenschtem alpha-Wert versehen:
   glDrawPixels(BildschirmBreite, BildschirmHoehe, GL_RGBA, GL_UNSIGNED_BYTE, NeuesBild.data());
   // Flushen, so dass das Bild in akzeptabler Verzoegerung auch sichtbar wird:
   glFinish();
   // Wenn Zyklus bei leistungsfaehiger Grafik zu schnell wuerde, dann muss hier gebremst werden:
   Timer.UeberblendungsZyklusVerzoegern();
  }
}  // XBildschirmUeberblenden


/*
 * PickSortProXOpenGL.cpp ************************************************************************************
 */

